(* %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
Routine solving the coupled Boltzmann equations (cBE):

    Y'=-(Subscript[\[Lambda], \[Chi]]Y^2-\[Lambda]Yeq^2),
    y'= [Eq.(28) in 1706.07433 or with full Cel]

Input as global function:

    svx[x]  -   thermally averaged annihilation cross section  < \[Sigma]v >
    sv2x[x] -   weighted annihilation cross section  < \[Sigma]v>_2
    gam[x]  - momentum transfer rate \[Gamma] (if FullCel = False in settings.wl)
    or GetSecondMomentScat[ ... ] - second momentum moment of Cel (if FullCel=True in settings.wl)

Output:

    tYcBE   -   table containing Y(x)
    tycBE   -   table containing y(x)
    Oh2cBE  -   final value of relic abundance

Version: 15-01-2021
Contact: Tobias Binder (Kavli IPMU)
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%% *)

cBE:= Module[{xi,xf,h,hh,eps,errNewton,errMaxNewton,safety,err,KDo,xip1,Yi,Yip1T,
              Yip1E,yi,yip1,yip1T,yip1E,Yeqi,Yeqip1,yeqi,yeqip1,lambdai,
              lambdaip1,lambdachii,c,u,cc,uchi,rhsi,rhsTip1,rhsEip1,lli,llip1,
              gamDsip1,svxip1,sv2xip1,TJacobi,EJacobi,delta,scat,scatpref,
              NewtonError,xDMip1,svxDMip1,sv2xDMip1,sq,prod1,prod2,
              DersvxDMip1,Dersv2xDMip1,DerYip1T,DerYip1E,w34,onepgt,
              Again,End,NewtonTrapezoidal,NewtonEuler,NewtonTrapezoidalKDONLY
              ,NewtonEulerKDONLY,tcBE},

    xi                              =   xmin+0.;
    xf                              =   xmax+0.;
    h                               =   cBEhinit+0.;
    eps                             =   cBEerr+0.;
    errNewton                       =   cBEerrNewton+0.;
    errMaxNewton                    =   cBEerrMaxNewton+0.;
    safety                          =   cBEsafety+0.;
    KDo                             =   KDonly;

    (* Initialize *)

    Yeqi                            =   Yeq[mDM,xi];
    Yi                              =   Yeqi;
    yi                              =   yeq[mDM,xi];
    lli                             =   s[mDM/xi]/(H[mDM/xi]xi)(1+gtilde[mDM/xi]);
    lambdai                         =   lli*svx[xi];
    lambdachii                      =   lambdai;
    rhsi                            =   0.;
    tcBE                            =   {{xi,Yi,yi}};

    While[xi<xf,

    hh                              =   h;
    If[xi+hh>xf,hh=xf-xi];

    Label[Again];

        xip1                        =   xi+hh;

        If[xip1==xi,Print["cBE.wl: error, stepsize overflow"];Goto[End]];

        If[!KDo,

            (********************************* cBE **************************************)

            onepgt                  =   (1+gtilde[mDM/xip1]);
            llip1                   =   s[mDM/xip1]/(H[mDM/xip1]xip1)onepgt;
            svxip1                  =   svx[xip1];
            lambdaip1               =   llip1*svxip1;

            If[lambdaip1==0,{Print["cBE.wl: svx zero at ", xip1,". Keeping Y_i constant, continuing with KD only."],KDo=True,Goto[Again]}];

            Yeqip1                  =   Yeq[mDM,xip1];
            u                       =   hh*lambdaip1;
            c                       =   2.*Yi+u*((Yeqip1^2+(lambdai/lambdaip1)*Yeqi^2)-(lambdachii/lambdaip1)*Yi^2);
            cc                      =   4.*(Yi+u*Yeqip1^2);
            yeqip1                  =   yeq[mDM,xip1];
            sv2xip1                 =   sv2x[xip1];

            If[!FullCel,gamDsip1=gam[xip1]/s[mDM/xip1],scatpref=(mDM^3/(H[mDM/xip1]s[mDM/xip1]^(2/3)xip1))onepgt];

            (* Implicit Trapezoidal method *)

            yip1                    =   yi;

            Label[NewtonTrapezoidal];

                xDMip1              =   xip1*yeqip1/yip1;
                svxDMip1            =   svx[xDMip1];
                sv2xDMip1           =   sv2x[xDMip1];
                uchi                =   hh*lambdaip1*svxDMip1/svxip1;

                If[1.+uchi*c<0.,hh=0.5*hh;Goto[Again]];

                sq                  =   Sqrt[1.+uchi*c];
                Yip1T               =   c/(1.+sq);
                prod1               =   llip1*yip1*Yip1T;
                prod2               =   (svxDMip1-sv2xDMip1-(Yeqip1/Yip1T)^2(svxip1-yeqip1/yip1 sv2xip1));
                w34                 =   GetRelTempCorrection[xDMip1,ExpDbk2[xDMip1],Derbk2Dbk2[xDMip1]];
                DersvxDMip1         =   -(xDMip1/yip1)*dsvx[xDMip1];
                Dersv2xDMip1        =   -(xDMip1/yip1)*dsv2x[xDMip1];
                DerYip1T            =   -Yip1T^2(hh*lambdaip1*DersvxDMip1)/(2.*svxip1*sq);

                If[!FullCel,
                    (* Fokker-Planck approximation *)
                    rhsTip1         =   onepgt*(yip1/xip1)w34[[1]]+llip1*gamDsip1*(1.-w34[[1]]/2.)(yeqip1-yip1)+prod1*prod2;
                    TJacobi         =   (onepgt*(1./xip1)(w34[[1]]-xDMip1 w34[[2]]))-(gamDsip1*(1.-w34[[1]]/2.)llip1)+ llip1*gamDsip1*((xDMip1/yip1)w34[[2]]/2.)(yeqip1-yip1)+prod1*(((1./yip1)+(DerYip1T/Yip1T))*prod2+(DersvxDMip1-Dersv2xDMip1+2.(Yeqip1/Yip1T)^2(DerYip1T/Yip1T)(svxip1-yeqip1/yip1 sv2xip1)-(Yeqip1/Yip1T)^2 (sv2xip1 yeqip1/(yip1^2))))
                    ,
                    (* Cel *)
                    scat            =   scatpref*GetSecondMomentScat[xip1,yeqip1,yip1,ExpDbk2[xDMip1],Derbk2Dbk2[xDMip1],modelparameters];
                    rhsTip1         =   onepgt*(yip1/xip1)w34[[1]]+scat[[1]]+prod1*prod2;
                    TJacobi         =   (onepgt*(1./xip1)(w34[[1]]-xDMip1 w34[[2]]))+scat[[2]]+ prod1*(((1./yip1)+(DerYip1T/Yip1T))*prod2+(DersvxDMip1-Dersv2xDMip1+2.(Yeqip1/Yip1T)^2(DerYip1T/Yip1T)(svxip1-yeqip1/yip1 sv2xip1)-(Yeqip1/Yip1T)^2(sv2xip1 yeqip1/(yip1^2))))
                ];

                delta               =   (yip1-yi-(hh/2.)(rhsi+rhsTip1))/(1.-(hh/2.)(TJacobi));
                yip1                =   yip1-delta;
                NewtonError         =   Abs[(delta)/(yip1)];

                If[NewtonError>errMaxNewton,hh=hh/5.;Goto[Again]];
                If[NewtonError>errNewton,Goto[NewtonTrapezoidal]];

            yip1T                   =   yip1;
            xDMip1                  =   xip1*yeqip1/yip1;
            svxDMip1                =   svx[xDMip1];
            Yip1T                   =   (c/(1.+Sqrt[1.+(hh*lambdaip1*svxDMip1/svxip1)c]));
            w34                     =   GetRelTempCorrection[xDMip1,ExpDbk2[xDMip1],Derbk2Dbk2[xDMip1]];

            If[!FullCel,
                rhsTip1             =   onepgt*(yip1/xip1)w34[[1]]+llip1*gamDsip1(1.-w34[[1]]/2.)(yeqip1-yip1)+llip1 yip1 Yip1T*(svxDMip1-sv2x[xDMip1]-(Yeqip1/(Yip1T))^2 (svxip1 -yeqip1/yip1 sv2xip1))
                ,
                scat                =   scatpref*GetSecondMomentScat[xip1,yeqip1,yip1,ExpDbk2[xDMip1],Derbk2Dbk2[xDMip1],modelparameters];
                rhsTip1             =   onepgt*(yip1/xip1)w34[[1]]+scat[[1]]+llip1 yip1 Yip1T*(svxDMip1-sv2x[xDMip1]-(Yeqip1/(Yip1T))^2(svxip1-yeqip1/yip1 sv2xip1))
            ];

            (* Implicit Euler method *)

            yip1                    =   yi;

            Label[NewtonEuler];

                xDMip1              =   xip1*yeqip1/yip1;
                svxDMip1            =   svx[xDMip1];
                sv2xDMip1           =   sv2x[xDMip1];
                uchi                =   hh*lambdaip1*svxDMip1/svxip1;

                If[1.+uchi*cc<0.,hh =0.5*hh;Goto[Again]];

                sq                  =   Sqrt[1.+uchi*cc];
                Yip1E               =   0.5*cc/(1.+sq);
                prod1               =   llip1*yip1*Yip1E;
                prod2               =   (svxDMip1-sv2xDMip1-(Yeqip1/Yip1E)^2 (svxip1-yeqip1/yip1 sv2xip1));
                w34                 =   GetRelTempCorrection[xDMip1,ExpDbk2[xDMip1],Derbk2Dbk2[xDMip1]];
                DersvxDMip1         =   -(xDMip1/yip1)*dsvx[xDMip1];
                Dersv2xDMip1        =   -(xDMip1/yip1)*dsv2x[xDMip1];
                DerYip1E            =   -Yip1E^2(hh*lambdaip1*DersvxDMip1)/(svxip1*sq);

                If[!FullCel,
                    rhsEip1         =   onepgt*(yip1/xip1)w34[[1]]+llip1*gamDsip1*(1.-w34[[1]]/2.)*(yeqip1-yip1)+prod1*prod2;
                    EJacobi         =   (onepgt*(1./xip1)(w34[[1]]-xDMip1 w34[[2]]))-(gamDsip1*(1.-w34[[1]]/2.)llip1)+llip1*gamDsip1*((xDMip1/yip1)w34[[2]]/2.)(yeqip1-yip1)+prod1*(((1./yip1)+(DerYip1E/Yip1E))*prod2+(DersvxDMip1-Dersv2xDMip1+2.(Yeqip1/Yip1E)^2(DerYip1E/Yip1E)(svxip1-yeqip1/yip1 sv2xip1)-(Yeqip1/Yip1E)^2(sv2xip1 yeqip1/(yip1^2))))
                    ,
                    scat            =   scatpref*GetSecondMomentScat[xip1,yeqip1,yip1,ExpDbk2[xDMip1],Derbk2Dbk2[xDMip1],modelparameters];
                    rhsEip1         =   onepgt*(yip1/xip1)w34[[1]]+scat[[1]]+prod1*prod2;
                    EJacobi         =   (onepgt*(1./xip1)(w34[[1]]-xDMip1 w34[[2]]))+scat[[2]]+prod1*(((1./yip1)+(DerYip1E/Yip1E))*prod2+(DersvxDMip1-Dersv2xDMip1+2.(Yeqip1/Yip1E)^2(DerYip1E/Yip1E)(svxip1-yeqip1/yip1 sv2xip1)-(Yeqip1/Yip1E)^2(sv2xip1 yeqip1/(yip1^2))))
                ];

                delta               =   (yip1-yi-(hh)(rhsEip1))/(1.-(hh)(EJacobi));
                yip1                =   yip1-delta;
                NewtonError         =   Abs[(delta)/(yip1)];

                If[NewtonError>errMaxNewton,hh=hh/5.;Goto[Again]];

                If[NewtonError>errNewton,Goto[NewtonEuler]];

            yip1E                   =   yip1;
            Yip1E                   =   0.5*(cc/(1.+Sqrt[1.+(hh*lambdaip1*svx[xip1*yeqip1/yip1]/svxip1)cc]));
            err                     =   Max[Abs[(Yip1T-Yip1E)/(Yip1T*eps)],Abs[(yip1T-yip1E)/(yip1T*eps)]];

            If[err>1,hh=Max[safety/Sqrt[err],0.1]*hh;Goto[Again]];

            xi                      =   xip1;
            Yi                      =   Yip1T;
            yi                      =   yip1T;
            Yeqi                    =   Yeqip1;
            yeqi                    =   yeqip1;
            lambdai                 =   lambdaip1;
            lambdachii              =   (lambdaip1*svx[xi*yeqi/yi]/svxip1);
            rhsi                    =   rhsTip1;

            ,

            (********************************** KD only ***********************************)

            onepgt                  =   (1+gtilde[mDM/xip1]);
            llip1                   =   s[mDM/xip1]/(H[mDM/xip1] xip1)onepgt;
            yeqip1                  =   yeq[mDM,xip1];

            If[!FullCel,gamDsip1=gam[xip1]/s[mDM/xip1],scatpref=(mDM^3/(H[mDM/xip1]s[mDM/xip1]^(2/3)xip1))onepgt];

            (* Implicit Trapezoidal method *)

            yip1                    =   yi;

            Label[NewtonTrapezoidalKDONLY];

                xDMip1              =   xip1*yeqip1/yip1;
                w34                 =   GetRelTempCorrection[xDMip1,ExpDbk2[xDMip1],Derbk2Dbk2[xDMip1]];

                If[!FullCel,
                    rhsTip1         =   onepgt*(yip1/xip1)w34[[1]]+llip1*gamDsip1*(1.-w34[[1]]/2.)(yeqip1-yip1);
                    TJacobi         =   (onepgt*(1./xip1)(w34[[1]]-xDMip1 w34[[2]]))-(gamDsip1*(1.-w34[[1]]/2.)llip1)+ llip1*gamDsip1*((xDMip1/yip1)w34[[2]]/2.)(yeqip1-yip1)
                    ,
                    scat            =   scatpref*GetSecondMomentScat[xip1,yeqip1,yip1,ExpDbk2[xDMip1],Derbk2Dbk2[xDMip1],modelparameters];
                    rhsTip1         =   onepgt*(yip1/xip1)w34[[1]]+scat[[1]];
                    TJacobi         =   (onepgt*(1./xip1)(w34[[1]]-xDMip1 w34[[2]]))+scat[[2]]
                ];

                delta               =   (yip1-yi-(hh/2.)(rhsi+rhsTip1))/(1.-(hh/2.)(TJacobi));
                yip1                =   yip1-delta;
                NewtonError         =   Abs[(delta)/(yip1)];

                If[NewtonError>errMaxNewton,hh = hh/5.;Goto[Again]];

                If[NewtonError>errNewton,Goto[NewtonTrapezoidalKDONLY]];

            yip1T                   =   yip1;
            xDMip1                  =   xip1*yeqip1/yip1;
            w34                     =   GetRelTempCorrection[xDMip1,ExpDbk2[xDMip1],Derbk2Dbk2[xDMip1]];

            If[!FullCel,
                rhsTip1             =   onepgt*(yip1/xip1)w34[[1]]+llip1*gamDsip1(1.-w34[[1]]/2.)(yeqip1-yip1)
                ,
                scat                =   scatpref*GetSecondMomentScat[xip1,yeqip1,yip1,ExpDbk2[xDMip1],Derbk2Dbk2[xDMip1],modelparameters];
                rhsTip1             =   onepgt*(yip1/xip1)w34[[1]]+scat[[1]]
            ];

            (* Implicit Euler method *)

            yip1                    =   yi;

            Label[NewtonEulerKDONLY];

                xDMip1              =   xip1*yeqip1/yip1;
                w34                 =   GetRelTempCorrection[xDMip1,ExpDbk2[xDMip1],Derbk2Dbk2[xDMip1]];

                If[!FullCel,
                    rhsEip1         =   onepgt*(yip1/xip1)w34[[1]]+llip1*gamDsip1*(1.-w34[[1]]/2.)*(yeqip1-yip1);
                    EJacobi         =   (onepgt*(1./xip1)(w34[[1]]-xDMip1 w34[[2]]))-(gamDsip1*(1.-w34[[1]]/2.)llip1)+ llip1*gamDsip1*((xDMip1/yip1)w34[[2]]/2.)(yeqip1-yip1)
                    ,
                    scat            =   scatpref*GetSecondMomentScat[xip1,yeqip1,yip1,ExpDbk2[xDMip1],Derbk2Dbk2[xDMip1],modelparameters];
                    rhsEip1         =   onepgt*(yip1/xip1)w34[[1]]+scat[[1]];
                    EJacobi         =   (onepgt*(1./xip1)(w34[[1]]-xDMip1 w34[[2]]))+scat[[2]]
                ];

                delta               =   (yip1-yi-(hh)(rhsEip1))/(1.-(hh)(EJacobi));
                yip1                =   yip1-delta;
                NewtonError         =   Abs[(delta)/(yip1)];

                If[NewtonError>errMaxNewton,hh = hh/5.;Goto[Again]];

                If[NewtonError>errNewton,Goto[NewtonEulerKDONLY]];

            yip1E                   =   yip1;
            err                     =   Abs[(yip1T-yip1E)/(yip1T*eps)];

            If[err>1,hh=Max[safety/Sqrt[err], 0.1]*hh; Goto[Again]];

            xi                      =   xip1;
            yi                      =   yip1T;
            yeqi                    =   yeqip1;
            rhsi                    =   rhsTip1;

        ]

        AppendTo[tcBE,{xi,Yi,yi}];
        If[err==0,h=5.*hh,h=Min[safety/Sqrt[err], 5.0]*hh];

    ];

    Label[End];

    tYcBE                           =   tcBE[[All,{1, 2}]];
    tycBE                           =   tcBE[[All,{1, 3}]];
    Oh2cBE                          =   mDM*Entropytoday/Rhocritical*Last[tYcBE][[2]];

    (* result for the relic density *)
    "Oh2cBE = " <> ToString[Oh2cBE]
];
